@@ -13,25 +13,34 @@ use ndarray::RemoveAxis;
 use ndarray::Zip;
 use ndarray::{stack, Array, Dimension};
 use ndarray::{ArrayView, ArrayViewMut, ArrayViewMut1};
+use num::Float;
+use num::Zero;
+use ordered_float::OrderedFloat;
 use rand::distributions::Uniform;
 use rand::rngs::StdRng;
+use serde::{Deserialize, Serialize};
+use std::cmp;
 use std::fmt::Display;
 use std::ops::{Mul, MulAssign};
 
-pub type Bounds1<T> = Bounds<T, Ix2>;
+pub type Bounds1 = Bounds<Ix2>;
 
-#[derive(Clone, Debug, PartialEq)]
-pub struct Bounds<T: NNVFloat, D: Dimension> {
-    data: Array<T, D>,
+#[derive(Clone, Debug, PartialEq, Deserialize, Serialize)]
+pub struct Bounds<D: Dimension> {
+    data: Array<NNVFloat, D>,
 }
 
-impl<T: NNVFloat, D: Dimension + ndarray::RemoveAxis> Bounds<T, D> {
+impl<D: Dimension + ndarray::RemoveAxis> Bounds<D> {
     /// # Panics
     pub fn new<'a, S: Dimension + Dimension<Larger = D>>(
-        lower: ArrayView<'a, T, S>,
-        upper: ArrayView<'a, T, S>,
+        lower: ArrayView<'a, NNVFloat, S>,
+        upper: ArrayView<'a, NNVFloat, S>,
     ) -> Self {
-        let data: Array<T, D> = stack(Axis(0), &[lower, upper]).unwrap();
+        debug_assert!(
+            Zip::from(&lower).and(&upper).all(|&l, &u| l <= u),
+            "Input bounds are flipped!"
+        );
+        let data: Array<NNVFloat, D> = stack(Axis(0), &[lower, upper]).unwrap();
         Self { data }
     }
 
@@ -41,33 +50,33 @@ impl<T: NNVFloat, D: Dimension + ndarray::RemoveAxis> Bounds<T, D> {
             .map_collect(|&lb, &ub| lb == ub)
     }
 
-    pub fn fixed_vals_or_zeros(&self) -> Array<T, D::Smaller> {
+    pub fn fixed_vals_or_zeros(&self) -> Array<NNVFloat, D::Smaller> {
         Zip::from(self.lower())
             .and(self.upper())
-            .map_collect(|&lb, &ub| if lb == ub { lb } else { T::zero() })
+            .map_collect(|&lb, &ub| if lb == ub { lb } else { NNVFloat::zero() })
     }
 
     pub fn is_all_finite(&self) -> bool {
-        self.data.iter().all(|&x| T::is_finite(x))
+        self.data.iter().all(|&x| NNVFloat::is_finite(x))
     }
 
-    pub fn as_tuple(&self) -> (Array<T, D::Smaller>, Array<T, D::Smaller>) {
+    pub fn as_tuple(&self) -> (Array<NNVFloat, D::Smaller>, Array<NNVFloat, D::Smaller>) {
         (self.lower().to_owned(), self.upper().to_owned())
     }
 
-    pub fn lower(&self) -> ArrayView<T, D::Smaller> {
+    pub fn lower(&self) -> ArrayView<NNVFloat, D::Smaller> {
         self.data.index_axis(Axis(0), 0)
     }
 
-    pub fn lower_mut(&mut self) -> ArrayViewMut<T, D::Smaller> {
+    pub fn lower_mut(&mut self) -> ArrayViewMut<NNVFloat, D::Smaller> {
         self.data.index_axis_mut(Axis(0), 0)
     }
 
-    pub fn upper(&self) -> ArrayView<T, D::Smaller> {
+    pub fn upper(&self) -> ArrayView<NNVFloat, D::Smaller> {
         self.data.index_axis(Axis(0), 1)
     }
 
-    pub fn upper_mut(&mut self) -> ArrayViewMut<T, D::Smaller> {
+    pub fn upper_mut(&mut self) -> ArrayViewMut<NNVFloat, D::Smaller> {
         self.data.index_axis_mut(Axis(0), 1)
     }
 
@@ -75,50 +84,68 @@ impl<T: NNVFloat, D: Dimension + ndarray::RemoveAxis> Bounds<T, D> {
         self.data.shape().iter().skip(1).product()
     }
 
-    pub fn bounds_iter(&self) -> Lanes<T, D::Smaller> {
+    pub fn bounds_iter(&self) -> Lanes<NNVFloat, D::Smaller> {
         self.data.lanes(Axis(0))
     }
 
-    pub fn bounds_iter_mut(&mut self) -> LanesMut<T, D::Smaller> {
+    pub fn bounds_iter_mut(&mut self) -> LanesMut<NNVFloat, D::Smaller> {
         self.data.lanes_mut(Axis(0))
     }
 
-    pub fn is_member(&self, x: &ArrayView<T, D::Smaller>) -> bool {
-        let eps = (1e-5).into();
+    pub fn is_member(&self, x: &ArrayView<NNVFloat, D::Smaller>) -> bool {
+        let eps = 1e-5;
         Zip::from(x)
             .and(self.bounds_iter())
             .all(|&x, bounds| bounds[0] - eps <= x && x <= bounds[1] + eps)
     }
+
+    #[must_use]
+    pub fn intersect(&self, other: &Self) -> Self {
+        let mut intersection = Self {
+            data: self.data.clone(),
+        };
+        Zip::from(self.lower())
+            .and(other.lower())
+            .map_assign_into(intersection.lower_mut(), |&x, &y| {
+                cmp::max(OrderedFloat(x), OrderedFloat(y)).0
+            });
+        Zip::from(self.upper())
+            .and(other.upper())
+            .map_assign_into(intersection.upper_mut(), |&x, &y| {
+                cmp::min(OrderedFloat(x), OrderedFloat(y)).0
+            });
+        intersection
+    }
 }
 
-impl<T: crate::NNVFloat, D: Dimension + ndarray::RemoveAxis> Bounds<T, D> {
+impl<D: Dimension + ndarray::RemoveAxis> Bounds<D> {
     /// # Panics
     pub fn subset(&self, rhs: &Self) -> bool {
         Zip::from(self.bounds_iter())
             .and(rhs.bounds_iter())
             .all(|me, rhs| {
                 let diff = me.to_owned() - rhs;
-                let eps = <T as num::NumCast>::from(1e-8).unwrap();
-                (diff[[0]] >= T::zero() || diff[[0]] <= eps)
-                    && (diff[[1]] <= T::zero() || diff[[1]] <= eps)
+                let eps = <NNVFloat as num::NumCast>::from(1e-8).unwrap();
+                (diff[[0]] >= NNVFloat::zero() || diff[[0]] <= eps)
+                    && (diff[[1]] <= NNVFloat::zero() || diff[[1]] <= eps)
             })
     }
 }
 
-impl<T: NNVFloat, D: Dimension + ndarray::RemoveAxis> Bounds<T, D> {
-    pub fn sample_uniform(&self, seed: u64) -> Array<T, D::Smaller> {
+impl<D: Dimension + ndarray::RemoveAxis> Bounds<D> {
+    pub fn sample_uniform(&self, seed: u64) -> Array<NNVFloat, D::Smaller> {
         let mut rng = StdRng::seed_from_u64(seed);
         Zip::from(self.bounds_iter())
             .map_collect(|x| Uniform::new_inclusive(x[0], x[1]).sample(&mut rng))
     }
 }
 
-impl<T: NNVFloat> Bounds1<T> {
+impl Bounds1 {
     /// # Panics
-    pub fn new_by_dim(dim_bounds: &[ArrayView1<T>]) -> Self {
-        let dims: Vec<_> = dim_bounds.iter().map(|x| x.insert_axis(Axis(0))).collect();
+    pub fn new_by_dim(dim_bounds: &[ArrayView1<NNVFloat>]) -> Self {
+        let dims: Vec<_> = dim_bounds.iter().map(|x| x.insert_axis(Axis(1))).collect();
         Self {
-            data: concatenate(Axis(0), &dims).unwrap(),
+            data: concatenate(Axis(1), &dims).unwrap(),
         }
     }
 
@@ -129,13 +156,15 @@ impl<T: NNVFloat> Bounds1<T> {
     }
 
     pub fn trivial(dim: usize) -> Self {
+        panic!("Trying to banish infinity from the library inputs.");
         Self::new(
-            Array::from_elem(dim, T::neg_infinity()).view(),
-            Array::from_elem(dim, T::infinity()).view(),
+            Array::from_elem(dim, NNVFloat::neg_infinity()).view(),
+            Array::from_elem(dim, NNVFloat::infinity()).view(),
         )
     }
 
-    pub fn affine_map(&self, aff: &Affine2<T>) -> Self {
+    #[must_use]
+    pub fn affine_map(&self, aff: &Affine2) -> Self {
         let lower = aff.apply(&self.lower());
         let upper = aff.apply(&self.upper());
         Self::new(lower.view(), upper.view())
@@ -154,15 +183,17 @@ impl<T: NNVFloat> Bounds1<T> {
     }
 
     /// # Panics
+    #[must_use]
     pub fn append(mut self, other: &Self) -> Self {
         self.data.append(Axis(1), other.data.view()).unwrap();
         self
     }
 
-    pub fn index_mut(&mut self, index: usize) -> ArrayViewMut1<T> {
+    pub fn index_mut(&mut self, index: usize) -> ArrayViewMut1<NNVFloat> {
         self.data.index_axis_mut(Axis(1), index)
     }
 
+    #[must_use]
     pub fn get_ith_bounds(&self, index: usize) -> Self {
         Self {
             data: self
@@ -175,10 +206,10 @@ impl<T: NNVFloat> Bounds1<T> {
 }
 
 /// Scale by scalar
-impl<T: NNVFloat, D: Dimension> Mul<T> for Bounds<T, D> {
+impl<D: Dimension> Mul<NNVFloat> for Bounds<D> {
     type Output = Self;
 
-    fn mul(self, rhs: T) -> Self {
+    fn mul(self, rhs: NNVFloat) -> Self {
         Self {
             data: self.data * rhs,
         }
@@ -186,13 +217,13 @@ impl<T: NNVFloat, D: Dimension> Mul<T> for Bounds<T, D> {
 }
 
 /// Scale by scalar
-impl<T: NNVFloat, D: Dimension> MulAssign<T> for Bounds<T, D> {
-    fn mul_assign(&mut self, rhs: T) {
+impl<D: Dimension> MulAssign<NNVFloat> for Bounds<D> {
+    fn mul_assign(&mut self, rhs: NNVFloat) {
         self.data *= rhs;
     }
 }
 
-impl<T: NNVFloat, D: Dimension + RemoveAxis> Display for Bounds<T, D> {
+impl<D: Dimension + RemoveAxis> Display for Bounds<D> {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::result::Result<(), std::fmt::Error> {
         write!(f, "Lower: {}\nUpper: {}", self.lower(), self.upper())
     }
