@@ -4,22 +4,24 @@
 use crate::affine::{Affine, Affine2, Affine4};
 use crate::bounds::Bounds1;
 use crate::gaussian::GaussianDistribution;
-use crate::inequality::Inequality;
 use crate::lp::solve;
 use crate::lp::LinearSolution;
 use crate::polytope::Polytope;
 use crate::tensorshape::TensorShape;
 use crate::NNVFloat;
-use ndarray::Array4;
+use ndarray::array;
 use ndarray::ArrayView1;
 use ndarray::Dimension;
 use ndarray::Ix4;
 use ndarray::{Array1, Array2};
+use ndarray::{Array4, ArrayView2};
 use ndarray::{Axis, Ix2};
+use num::Float;
+use serde::{Deserialize, Serialize};
 use std::fmt::Debug;
 
-pub type Star2<A> = Star<A, Ix2>;
-pub type Star4<A> = Star<A, Ix4>;
+pub type Star2 = Star<Ix2>;
+pub type Star4 = Star<Ix4>;
 
 /// Representation of a set acted on by a deep neural network (DNN)
 ///
@@ -39,36 +41,36 @@ pub type Star4<A> = Star<A, Ix4>;
 /// Based on: Tran, Hoang-Dung, et al. "Star-based reachability analysis of
 /// deep neural networks." International Symposium on Formal Methods. Springer,
 /// Cham, 2019.
-#[derive(Clone, Debug)]
-pub struct Star<T: NNVFloat, D: Dimension> {
+#[derive(Clone, Debug, Deserialize, Serialize)]
+pub struct Star<D: Dimension> {
     /// `representation` is the concatenation of [basis center] (where
     /// center is a column vector) and captures information about the
     /// transformed set
-    representation: Affine<T, D>,
+    representation: Affine<D>,
     /// `constraints` is the concatenation of [coeffs upper_bounds]
     /// and is a representation of the input polyhedron
-    constraints: Option<Polytope<T>>,
+    constraints: Option<Polytope>,
 }
 
-impl<T: NNVFloat, D: Dimension> Star<T, D> {
+impl<D: Dimension> Star<D> {
     pub fn ndim(&self) -> usize {
         self.representation.ndim()
     }
 
-    pub fn input_space_polytope(&self) -> Option<&Polytope<T>> {
+    pub fn input_space_polytope(&self) -> Option<&Polytope> {
         self.constraints.as_ref()
     }
 
-    pub fn center(&self) -> ArrayView1<T> {
+    pub fn center(&self) -> ArrayView1<NNVFloat> {
         self.representation.shift()
     }
 
-    pub fn get_representation(&self) -> &Affine<T, D> {
+    pub fn get_representation(&self) -> &Affine<D> {
         &self.representation
     }
 }
 
-impl<T: NNVFloat, D: Dimension> Star<T, D> {
+impl<D: Dimension> Star<D> {
     pub fn num_constraints(&self) -> usize {
         match &self.constraints {
             Some(polytope) => polytope.num_constraints(),
@@ -78,44 +80,61 @@ impl<T: NNVFloat, D: Dimension> Star<T, D> {
 
     /// Add constraints to restrict the input set. Each row represents a
     /// constraint and the last column represents the upper bounds.
-    pub fn add_constraints(
-        mut self,
-        new_constraints: &Inequality<T>,
-        check_redundant: bool,
-    ) -> Self {
+    #[must_use]
+    pub fn add_constraint(mut self, coeffs: ArrayView1<NNVFloat>, rhs: NNVFloat) -> Self {
         if let Some(ref mut constrs) = self.constraints {
-            constrs.add_constraints(new_constraints, check_redundant);
+            constrs.add_eqn(coeffs, rhs);
         } else {
-            self.constraints = Some(Polytope::from_halfspaces(new_constraints.clone()));
+            self.constraints =
+                Polytope::nonempty_new(coeffs.to_owned().insert_axis(Axis(0)), array![rhs]);
+        }
+        self
+    }
+
+    pub fn remove_constraint(mut self, idx: usize) -> Self {
+        if let Some(ref mut constrs) = self.constraints {
+            constrs.remove_eqn(idx);
+        } else {
+            panic!();
         }
         self
     }
 }
 
-impl<T: NNVFloat> Star2<T> {
-    pub fn get_constraint_coeffs(&self) -> Option<Array2<T>> {
+impl Star2 {
+    pub fn get_constraint_coeffs(&self) -> Option<Array2<NNVFloat>> {
         self.constraints.as_ref().map(|x| x.coeffs().to_owned())
     }
 
-    pub fn get_safe_subset(&self, safe_value: T) -> Self {
-        let subset = self.clone();
-        let mut new_constr: Inequality<T> = self.representation.clone().into();
+    #[must_use]
+    pub fn get_safe_subset(&self, safe_value: NNVFloat) -> Self {
+        let mut subset = self.clone();
+        let mut new_constr: Polytope = self.representation.clone().into();
         let mut rhs = new_constr.rhs_mut();
-        rhs *= T::neg(T::one());
+        rhs *= -1.;
         rhs += safe_value;
-        subset.add_constraints(&new_constr, false)
+        subset.intersect_input(&new_constr);
+        subset
+    }
+
+    pub fn intersect_input(&mut self, other: &Polytope) {
+        self.constraints = self
+            .constraints
+            .as_ref()
+            .map_or(Some(other.clone()), |x| Some(x.intersect(other)));
     }
 
     pub fn get_input_trunc_gaussian(
         &self,
-        mu: &Array1<T>,
-        sigma: &Array2<T>,
+        mu: ArrayView1<NNVFloat>,
+        sigma: ArrayView2<NNVFloat>,
         max_accept_reject_iters: usize,
-        stability_eps: T,
-    ) -> Option<GaussianDistribution<T>> {
+        stability_eps: NNVFloat,
+        bounds_opt: &Option<Bounds1>,
+    ) -> Option<GaussianDistribution> {
         self.constraints
             .as_ref()
-            .and_then(Polytope::reduce_fixed_inputs)
+            .and_then(|x| x.reduce_fixed_inputs(bounds_opt))
             .map(|poly| {
                 poly.get_truncnorm_distribution(mu, sigma, max_accept_reject_iters, stability_eps)
             })
@@ -139,7 +158,7 @@ impl<T: NNVFloat> Star2<T> {
     /// Create a new Star with given basis vector and center.
     ///
     /// By default this Star covers the space because it has no constraints. To add constraints call `.add_constraints`.
-    pub fn new(basis: Array2<T>, center: Array1<T>) -> Self {
+    pub fn new(basis: Array2<NNVFloat>, center: Array1<NNVFloat>) -> Self {
         Self {
             representation: Affine2::new(basis, center),
             constraints: None,
@@ -147,8 +166,9 @@ impl<T: NNVFloat> Star2<T> {
     }
 
     /// # Panics
-    pub fn with_constraints(mut self, constraints: Polytope<T>) -> Self {
-        debug_assert!(!self.constraints.is_some(), "explicit panic");
+    #[must_use]
+    pub fn with_constraints(mut self, constraints: Polytope) -> Self {
+        debug_assert!(self.constraints.is_none(), "explicit panic");
         self.constraints = Some(constraints);
         self
     }
@@ -164,53 +184,38 @@ impl<T: NNVFloat> Star2<T> {
     }
 }
 
-impl<T: NNVFloat> Star2<T> {
-    pub fn get_input_bounds(&self) -> Option<&Bounds1<T>> {
-        self.constraints.as_ref().map(Polytope::get_bounds)
-    }
-
-    pub fn with_input_bounds(mut self, input_bounds: Bounds1<T>) -> Self {
-        if self.constraints.is_some() {
-            self.constraints = self.constraints.map(|x| x.with_input_bounds(input_bounds));
-        } else {
-            self.constraints = Some(Polytope::from(input_bounds));
-        }
-        self
-    }
-}
-
-impl<T: NNVFloat> Star2<T> {
+impl Star2 {
     /// Apply an affine transformation to the representation
-    pub fn affine_map2(&self, affine: &Affine<T, Ix2>) -> Self {
+    #[must_use]
+    pub fn affine_map2(&self, affine: &Affine<Ix2>) -> Self {
         Self {
             representation: affine * &self.representation,
             constraints: self.constraints.clone(),
         }
     }
-}
 
-impl<T: NNVFloat> Star2<T> {
-    pub fn step_relu2(&self, index: usize) -> (Option<Self>, Option<Self>) {
-        let neg_one: T = std::convert::From::from(-1.);
-
-        let mut new_constr: Inequality<T> = {
-            let mut aff = self.representation.get_eqn(index);
-            let neg_basis_part = &aff.basis() * neg_one;
-            aff.basis_mut().assign(&neg_basis_part);
-            aff.into()
+    pub fn step_relu2(
+        &self,
+        index: usize,
+        input_bounds_opt: &Option<Bounds1>,
+    ) -> (Option<Self>, Option<Self>) {
+        let (coeffs, shift) = {
+            let aff = self.representation.get_eqn(index);
+            let neg_basis_part: Array2<NNVFloat> = &aff.basis() * -1.;
+            let shift = aff.shift();
+            (neg_basis_part.row(0).to_owned(), shift[[0]])
         };
-        let upper_star = self.clone().add_constraints(&new_constr, true);
+        let upper_star = self.clone().add_constraint(coeffs.view(), shift);
 
-        new_constr *= neg_one;
-        let mut lower_star = self.clone().add_constraints(&new_constr, true);
+        let mut lower_star = self.clone().add_constraint((&coeffs * -1.).view(), -shift);
         lower_star.representation.zero_eqn(index);
 
-        let lower_star_opt = if lower_star.is_empty() {
+        let lower_star_opt = if lower_star.is_empty(input_bounds_opt) {
             None
         } else {
             Some(lower_star)
         };
-        let upper_star_opt = if upper_star.is_empty() {
+        let upper_star_opt = if upper_star.is_empty(input_bounds_opt) {
             None
         } else {
             Some(upper_star)
@@ -218,12 +223,16 @@ impl<T: NNVFloat> Star2<T> {
         (lower_star_opt, upper_star_opt)
     }
 
-    pub fn step_relu2_dropout(&self, index: usize) -> (Option<Self>, Option<Self>, Option<Self>) {
+    pub fn step_relu2_dropout(
+        &self,
+        index: usize,
+        input_bounds_opt: &Option<Bounds1>,
+    ) -> (Option<Self>, Option<Self>, Option<Self>) {
         let mut dropout_star = self.clone();
         dropout_star.representation.zero_eqn(index);
 
-        let stars = self.step_relu2(index);
-        let dropout_star_opt = if dropout_star.is_empty() {
+        let stars = self.step_relu2(index, input_bounds_opt);
+        let dropout_star_opt = if dropout_star.is_empty(input_bounds_opt) {
             None
         } else {
             Some(dropout_star)
@@ -231,26 +240,6 @@ impl<T: NNVFloat> Star2<T> {
         (dropout_star_opt, stars.0, stars.1)
     }
 
-    #[cfg(feature = "rayon")]
-    pub fn par_get_min(&self, idx: usize) -> T {
-        let eqn = self.representation.get_eqn(idx);
-
-        if let Some(ref poly) = self.constraints {
-            let solved = par_solve(
-                poly.coeffs().rows(),
-                poly.ubs(),
-                eqn.basis().index_axis(Axis(0), 0),
-            );
-            if let LinearSolution::Solution(_, val) = solved {
-                eqn.shift()[0] + val.into()
-            } else {
-                panic!()
-            }
-        } else {
-            T::neg_infinity()
-        }
-    }
-
     /// Calculates the minimum value of the equation at index `idx`
     /// given the constraints
     ///
@@ -266,20 +255,27 @@ impl<T: NNVFloat> Star2<T> {
     /// solve function. Currently this is way too hard to do, so we
     /// panic instead. We have an assumption that we start with a
     /// bounded box and therefore should never be unbounded.
-    pub fn get_min(&self, idx: usize) -> T {
+    pub fn get_output_min(&self, idx: usize, input_bounds: &Bounds1) -> NNVFloat {
         let eqn = self.representation.get_eqn(idx);
+        let shift = eqn.shift()[0];
 
         self.constraints.as_ref().map_or_else(
-            || T::neg_infinity(),
+            || {
+                crate::util::signed_dot(&eqn.basis(), &input_bounds.lower(), &input_bounds.upper())
+                    [[0]]
+                    + shift
+            },
             |poly| {
                 let solved = solve(
                     poly.coeffs().rows(),
-                    poly.ubs(),
+                    poly.rhs(),
                     eqn.basis().index_axis(Axis(0), 0),
-                    poly.get_bounds(),
+                    Some(input_bounds),
                 );
                 if let LinearSolution::Solution(_, val) = solved {
-                    eqn.shift()[0] + val.into()
+                    shift + val
+                } else if let LinearSolution::Unbounded(_) = solved {
+                    NNVFloat::neg_infinity()
                 } else {
                     panic!("Solution: {:?}", solved)
                 }
@@ -287,29 +283,6 @@ impl<T: NNVFloat> Star2<T> {
         )
     }
 
-    #[cfg(feature = "rayon")]
-    pub fn par_get_max(&self, idx: usize) -> T {
-        let neg_one: T = std::convert::From::from(-1.);
-        let mut eqn = self.representation.get_eqn(idx);
-        let shift = eqn.shift()[0];
-        eqn *= neg_one;
-
-        if let Some(ref poly) = self.constraints {
-            let solved = par_solve(
-                poly.coeffs().rows(),
-                poly.ubs(),
-                eqn.basis().index_axis(Axis(0), 0),
-            );
-            if let LinearSolution::Solution(_, val) = solved {
-                shift - val.into()
-            } else {
-                panic!()
-            }
-        } else {
-            T::infinity()
-        }
-    }
-
     /// Calculates the maximum value of the equation at index `idx`
     /// given the constraints
     ///
@@ -318,23 +291,27 @@ impl<T: NNVFloat> Star2<T> {
     ///
     /// # Panics
     /// TODO: Change output type to Option<T>
-    pub fn get_max(&self, idx: usize) -> T {
-        let neg_one: T = std::convert::From::from(-1.);
+    pub fn get_output_max(&self, idx: usize, input_bounds: &Bounds1) -> NNVFloat {
         let mut eqn = self.representation.get_eqn(idx);
         let shift = eqn.shift()[0];
-        eqn *= neg_one;
 
         self.constraints.as_ref().map_or_else(
-            || T::infinity(),
+            || {
+                crate::util::signed_dot(&eqn.basis(), &input_bounds.upper(), &input_bounds.lower())
+                    [[0]]
+                    + shift
+            },
             |poly| {
                 let solved = solve(
                     poly.coeffs().rows(),
-                    poly.ubs(),
-                    eqn.basis().index_axis(Axis(0), 0),
-                    poly.get_bounds(),
+                    poly.rhs(),
+                    eqn.basis().index_axis(Axis(0), 0).mapv(|x| x * -1.).view(),
+                    Some(input_bounds),
                 );
                 if let LinearSolution::Solution(_, val) = solved {
-                    shift - val.into()
+                    shift - val
+                } else if let LinearSolution::Unbounded(_) = solved {
+                    NNVFloat::infinity()
                 } else {
                     panic!()
                 }
@@ -342,21 +319,84 @@ impl<T: NNVFloat> Star2<T> {
         )
     }
 
-    pub fn calculate_axis_aligned_bounding_box(&self) -> Bounds1<T> {
-        let lbs = Array1::from_iter((0..self.representation_space_dim()).map(|x| self.get_min(x)));
-        let ubs = Array1::from_iter((0..self.representation_space_dim()).map(|x| self.get_max(x)));
+    //pub fn nearest_euclidean_neighbor(&self, origin: Array1<NNVFloat>) -> Array1<NNVFloat> {}
+
+    /// # Panics
+    pub fn can_maximize_output_idx(&self, class_idx: usize) -> bool {
+        let class_eqn = self.representation.get_eqn(class_idx);
+        let (class_coeffs, class_shift): (ArrayView1<NNVFloat>, NNVFloat) = {
+            (
+                class_eqn.basis().remove_axis(Axis(0)),
+                class_eqn.shift()[[0]],
+            )
+        };
+        let nvars = class_coeffs.len();
+
+        if self.constraints.is_none() {
+            return true; // We should have more sophisticated handling here, but this is a stopgap
+        }
+        let poly = self.constraints.as_ref().unwrap();
+        let block_coeffs = poly.coeffs();
+        let (A, b) = (block_coeffs.rows(), poly.rhs());
+        // Add a constraint for each output class not equal to the one being maximized
+        let mut coeffs = Vec::new();
+        let mut shifts = Vec::new();
+        for idx in 0..self.representation.shift().ndim() {
+            if idx == class_idx {
+                continue;
+            }
+            {
+                let (diff_coeffs, diff_shift) = {
+                    let other_class_eqn = self.representation.get_eqn(idx);
+                    (
+                        &other_class_eqn.basis().row(0) - &class_coeffs,
+                        class_shift - other_class_eqn.shift()[[0]],
+                    )
+                };
+                coeffs.push(diff_coeffs);
+                shifts.push(diff_shift);
+            }
+        }
+
+        let solve_a = A
+            .into_iter()
+            .chain(coeffs.iter().map(ndarray::ArrayBase::view));
+        let solved = solve(
+            solve_a,
+            b.iter().chain(shifts.iter()),
+            Array1::ones(nvars).view(),
+            Some(&Bounds1::trivial(nvars)),
+        );
+        matches!(
+            solved,
+            LinearSolution::Solution(..) | LinearSolution::Unbounded(..)
+        )
+    }
+
+    pub fn calculate_output_axis_aligned_bounding_box(
+        &self,
+        input_outer_bounds: &Bounds1,
+    ) -> Bounds1 {
+        let lbs = Array1::from_iter(
+            (0..self.representation_space_dim())
+                .map(|x| self.get_output_min(x, input_outer_bounds)),
+        );
+        let ubs = Array1::from_iter(
+            (0..self.representation_space_dim())
+                .map(|x| self.get_output_max(x, input_outer_bounds)),
+        );
         Bounds1::new(lbs.view(), ubs.view())
     }
 
     /// Check whether the Star set is empty.
-    pub fn is_empty(&self) -> bool {
+    pub fn is_empty(&self, input_bounds_opt: &Option<Bounds1>) -> bool {
         self.constraints
             .as_ref()
-            .map_or(false, crate::polytope::Polytope::is_empty)
+            .map_or(false, |x| x.is_empty(input_bounds_opt))
     }
 }
 
-impl<T: NNVFloat> Star4<T> {
+impl Star4 {
     /// Create a new Star with given dimension.
     ///
     /// By default this Star covers the space because it has no constraints. To add constraints call `.add_constraints`.
@@ -381,12 +421,12 @@ impl<T: NNVFloat> Star4<T> {
 
 #[cfg(test)]
 mod test {
-    use super::*;
-    use crate::test_util::{array2, empty_star, non_empty_star};
-    use ndarray::arr1;
-    use proptest::prelude::*;
-    use proptest::proptest;
-    use std::panic;
+    // use super::*;
+    // use crate::test_util::{array2, empty_star, non_empty_star};
+    // use ndarray::arr1;
+    // use proptest::prelude::*;
+    // use proptest::proptest;
+    // use std::panic;
 
     /*
         #[test]
@@ -421,47 +461,47 @@ mod test {
             */
 
             #[test]
-            fn test_get_min_feasible(star in non_empty_star(2,3)) {
+            fn test_get_output_min_feasible(star in non_empty_star(2,3)) {
                 prop_assert!(!star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
                 prop_assert!(!star.is_empty(), "Non empty star is empty");
                 let result = panic::catch_unwind(|| {
-                    star.get_min(0);
+                    star.get_output_min(0);
                 });
                 prop_assert!(result.is_ok(), "Calculating min resulted in panic for feasible star");
             }
 
             #[test]
-            fn test_get_min_infeasible(star in empty_star(2,1)) {
+            fn test_get_output_min_infeasible(star in empty_star(2,1)) {
                 prop_assert!(star.input_space_polytope().unwrap().is_empty(), "Polytope is not empty");
                 prop_assert!(star.is_empty(), "Empty star is not empty");
                 let result = panic::catch_unwind(|| {
-                    star.get_min(0)
+                    star.get_output_min(0)
                 });
-                prop_assert!(result.is_err(), "Infeasible star did not panic for get_min {:#?}", result);
+                prop_assert!(result.is_err(), "Infeasible star did not panic for get_output_min {:#?}", result);
             }
 
             #[test]
-            fn test_get_max_feasible(star in non_empty_star(2,3)) {
+            fn test_get_output_max_feasible(star in non_empty_star(2,3)) {
                 prop_assert!(!star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
                 prop_assert!(!star.is_empty(), "Non empty star is empty");
                 let result = panic::catch_unwind(|| {
-                    star.get_max(0);
+                    star.get_output_max(0);
                 });
                 prop_assert!(result.is_ok(), "Calculating min resulted in panic for feasible star");
             }
 
             #[test]
-            fn test_get_max_infeasible(star in empty_star(2,1)) {
+            fn test_get_output_max_infeasible(star in empty_star(2,1)) {
                 prop_assert!(star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
                 prop_assert!(star.is_empty(), "Empty star is not empty");
                 let result = panic::catch_unwind(|| {
-                    star.get_max(0)
+                    star.get_output_max(0)
                 });
-                prop_assert!(result.is_err(), "Infeasible star did not panic for get_min {:#?}", result);
+                prop_assert!(result.is_err(), "Infeasible star did not panic for get_output_min {:#?}", result);
             }
 
             #[test]
-            fn test_get_min_box_polytope(basis in array2(2, 2)) {
+            fn test_get_output_min_box_polytope(basis in array2(2, 2)) {
                 let num_dims = 2;
                 let box_bounds: Bounds1<f64> = Bounds1::new(Array1::from_elem(num_dims, -20.).view(), Array1::from_elem(num_dims, -20.).view());
                 let box_ineq = Inequality::new(Array2::zeros([1, num_dims]), Array1::zeros(1), box_bounds);
@@ -471,13 +511,13 @@ mod test {
                 let star = Star2::new(basis, center).with_constraints(poly);
 
                 let result = panic::catch_unwind(|| {
-                    star.get_min(0);
+                    star.get_output_min(0);
                 });
                 assert!(result.is_ok());
             }
 
             #[test]
-            fn test_get_max_box_polytope(basis in array2(2, 2)) {
+            fn test_get_output_max_box_polytope(basis in array2(2, 2)) {
                 let num_dims = 2;
                 let box_bounds: Bounds1<f64> = Bounds1::new(Array1::from_elem(num_dims, -20.).view(), Array1::from_elem(num_dims, -20.).view());
                 let box_ineq = Inequality::new(Array2::zeros([1, num_dims]), Array1::zeros(1), box_bounds);
@@ -487,7 +527,7 @@ mod test {
                 let star = Star2::new(basis, center).with_constraints(poly);
 
                 let result = panic::catch_unwind(|| {
-                    star.get_max(0);
+                    star.get_output_max(0);
                 });
                 assert!(result.is_ok());
             }
